import logging.handlers
import os


class FileHandler(logging.handlers.RotatingFileHandler):
    def __init__(self, path='/home/', filename='cranial.log'):
        if path[-1] != '/':
            path += '/'

        logging.handlers.RotatingFileHandler.__init__(self,
                                                      path + filename,
                                                      maxBytes=5242880,
                                                      backupCount=10)
        fmt = '%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s'
        fmt_date = '%Y-%m-%dT%T%Z'
        formatter = logging.Formatter(fmt, fmt_date)

        self.setFormatter(formatter)


class StreamingHandler(logging.StreamHandler):
    def __init__(self):
        logging.StreamHandler.__init__(self)
        fmt = '%(asctime)s - %(filename)s:%(lineno)d - %(levelname)s - %(message)s'
        fmt_date = '%Y-%m-%dT%T%Z'
        formatter = logging.Formatter(fmt, fmt_date)

        self.setFormatter(formatter)


def create(name, level, log_type='streaming', filename=None):
    """Create new logger

    :param name: name for logger
    :type name: str
    :param level:  logging level, e.g. logging.WARN
    :type level: int
    :param type
    :returns: logging object
    """
    log = logging.getLogger(name)
    log.setLevel(level)
    if not log.handlers:
        if log_type == 'streaming':
            log.addHandler(StreamingHandler())
        else:
            if filename is not None:
                log.addHandler(FileHandler(filename))
            else:
                log.addHandler(FileHandler())
    return log


def get(var='LOGLEVEL', level=None, name='cranial'):
    """ This is the recommended way to instantiate a logger if you
    don't care about the details.

    >>> log = get()
    >>> log.debug('spam')
    <BLANKLINE>
    >>> log.warning('Ni!') # doctest: +ELLIPSIS
    ... WARNING - Ni!
    """
    level = level or fallback(var)
    return create(name, level)


def fallback(var_name, fallback='WARN'):
    return os.environ.get(var_name,
                          os.environ.get('CRANIAL_LOGLEVEL', fallback))



import io
import tarfile
import gzip

def dir2bytes(dirname:str) -> bytes:
    """Make an in-memory gzip'd tar archive from a directory with minimal copying.

    # Test below effectively copies current dir to /tmp/doctest.

    >>> from tempfile import mkstemp, mkdtemp
    >>> dir = mkdtemp()
    >>> f = mkstemp(dir=dir)
    >>> bytes2dir(dir2bytes(dir), '/tmp/doctest')
    """

    with io.BytesIO() as buf:
        with tarfile.TarFile(mode='w', fileobj=buf) as tar:
            tar.add(dirname)
            outbytes = gzip.compress(buf.getbuffer())
    return outbytes


def bytes2dir(tarbytes: bytes, dirname = '.') -> None:
    """Take the bytes that make-up an gzip'd tar archive and decompress to disk."""
    fileobj = io.BytesIO(tarbytes)
    with tarfile.TarFile(fileobj=io.BytesIO(gzip.decompress(tarbytes))) as tar:
        tar.extractall(dirname)



import logging
import socket
from typing import Callable, List


def available_port():
    """ Not without race conditions; another process could claim the port
    between when you get the number and when you use it. This is intended for
    testing; production code should probably use known, fixed port numbers.
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('localhost', 0))
    addr, port = s.getsockname()
    s.close()
    return port


def update_ordered_list(old: List, new: List) -> List:
    """Returns a list of all items in new, retaining the order of old for the
    common items.

    >>> update_ordered_list([1,2,3], [4,3,1])
    [1, 3, 4]
    >>> update_ordered_list([1,2,3], [4,1,3])
    [1, 3, 4]
    >>> update_ordered_list([5,6,7], [4,1,3])
    [4, 1, 3]
    >>> update_ordered_list([6,5,7], [6,7,5])
    [6, 5, 7]
    """

    # Fast check.
    if sorted(old) == sorted(new):
        return old

    result = [x for x in old if x in new]
    for item in new:
        if item not in result:
            result.append(item)

    return result


def str2bucket(text: str, num_buckets: int) -> int:
    """Takes a string and uses a consistent hash to return an int selected
    evenly from the range 0 to (num_buckets - 1). Useful for 'sticky' load
    balancing on an arbitrary string key.

    >>> str2bucket('foo', 99)
    91
    >>> str2bucket('bar', 99)
    63
    >>> str2bucket('foo', 3)
    1
    >>> str2bucket('bar', 3)
    0
    >>> str2bucket('baz', 0)
    0
    """
    if num_buckets == 0:
        # Can't do modulo 0 or universe is destroyed.
        return 0
    from hashlib import md5
    return int(md5(text.encode()).hexdigest(), 16) % num_buckets


def dieIf(msg: str, fn: Callable, *args, **kwargs):
    """Wrap Exceptions in friendlier messages."""
    try:
        return fn(*args, **kwargs)
    except Exception as e:
        logging.error('%s. Function: %s; Args: %s; Keywords: %s',
                      e, fn.__name__, args, kwargs)
        raise Exception(msg)


def warnIf(msg: str, fn: Callable, *args, **kwargs):
    """Transform exceptions into warnings."""
    try:
        return fn(*args, **kwargs)
    except Exception as e:
        logging.warn(e)
        logging.warn('%s. Function: %s; Args: %s; Keywords: %s',
                     msg, fn.__name__, args, kwargs)
        raise e
        return None



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['cranial']

package_data = \
{'': ['*'], 'cranial': ['common/*']}

setup_kwargs = {
    'name': 'cranial-common',
    'version': '0.3.0',
    'description': 'Low-level utilities shared by various Cranial packages.',
    'long_description': None,
    'author': 'Matt Chapman, et al.',
    'author_email': 'Matt@NinjitsuWeb.com',
    'url': 'https://github.com/tribune/cranial-common',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)



"""
This module exists to support the 12-factor App approach to configuration,
while also supporting convenient CLIs. This provides a subset of the
functionality of the `click` module, so you probably don't need this if you're
already using click.

CLI options take precendence over Environment Variables.

We assume, but do not enforce, the use of the `docopt` module to handle
configuration via command-line arguments. Any dict[str, str] of config values
will do, if you prefer to use lower level argument parsing.


Typical usage:
    1. Read command-line args first, e.g., `opts = docopt.docopt(__doc__)`
    2. Read and store environment variables, with
      `config.load(opts, 'some_env_var_prefix')`.
    3. Throughout the application, use `config.get()` to get a dict of
      configuration values.
"""

from collections import OrderedDict
import importlib
import os
from typing import Any, Dict, Optional, Tuple, Union  # noqa
import urllib.parse

import yaml

ConfigValue = Union[str, bool, Dict[str, Any]]
ConfigStore = Dict[str, ConfigValue]

immutable_config_values = None  # type: Optional[ConfigStore]


def opts2env(opts: Dict[str, str], prefix: str) -> None:
    """ Takes opts produced by docopt and puts then in Environment variables
    with the given case-insensitive prefix. `__name__` might make a good
    prefix.

    >>> opts2env({'--monty': 'spam', '<WITCH>': True, 'duck': False, 'a': 'b'},
    ...          'ctest')
    >>> os.environ['CTEST_MONTY']
    'spam'
    >>> os.environ['CTEST_WITCH']
    '1'
    >>> os.environ.get('CTEST_DUCK', 'missing')
    'missing'
    >>> os.environ['CTEST_A']
    'b'
    """
    prefix = prefix.upper()
    for k, v in opts.items():
        name = prefix + '_' + k.upper().strip('-<> ').replace('-', '_')
        if v is not False and v is not None:
            os.environ[name] = '1' if v is True else str(v)


def parse_uri(s: str) -> Dict:
    """Parse URIs into factory parameters."""
    parse = urllib.parse.urlparse(str(s))
    # Most of the Python ecosystem breaks if you name a module 'http',
    # so we use httpget.
    mod = 'httpget' if parse.scheme == 'http' else parse.scheme

    d = {'module': mod,
         'address': parse.netloc,
         'endpoint': parse.path[1:]}  # type: Dict[str, Any]
    if parse.username:
        d['user'] = parse.username
    if parse.password:
        d['password'] = parse.password
    if parse.query:
        q = urllib.parse.parse_qs(parse.query)
        d.update({k: i[0] if len(i) < 2 else i
                  for k, i in q.items()})
    return d


def load_from_env(prefix: str) -> ConfigStore:
    """ Returns a dict of environment variables with the given case-insensitive
    prefix.

    >>> os.environ['CTEST_FOO'] = '0'
    >>> os.environ['CTEST_BAR'] = 'hello'
    >>> os.environ['CTEST_URI'] = 'kafka://host/ok?mode=hot'
    >>> conf = load_from_env('CtEsT')
    >>> conf['FOO']
    False
    >>> conf['BAR']
    'hello'
    >>> conf['URI'] == dict(\
    module='kafka', address='host', endpoint='ok', mode='hot')
    True
    >>> conf['URI_STR']
    'kafka://host/ok?mode=hot'
    """
    prefix = prefix.upper()
    config = {}  # type: ConfigStore
    k = ''  # type: str
    v = ''  # type: ConfigValue
    for k, v in os.environ.items():
        if not k.startswith(prefix + '_'):
            continue
        # @TODO Not sure this is the best place for this.
        if type(v) is str and '://' in str(v):
            fullkey = k.upper() + '_STR'
            config[fullkey.replace(prefix+'_', '')] = v
            v = parse_uri(v)  # type: ignore
        elif type(v) is str:
            # Make falsey things actually False.
            v = False if str(v).lower() in ['0', 'false'] else v
        config[k.replace(prefix + '_', '')] = v
    return config


def load(opts: Dict[str, str], prefix: str, fname=None) -> ConfigStore:
    """
    >>> o = {}
    >>> o['--foo'] = '0'
    >>> o['--bar'] = 'hello'
    >>> o['--uri'] = 'kafka://host/ok?mode=hot'
    >>> conf = load(o, 'CteSt')
    >>> conf['FOO']
    False
    >>> conf['BAR']
    'hello'
    >>> conf['URI'] == dict(\
    module='kafka', address='host', endpoint='ok', mode='hot')
    True
    >>> os.environ['CTEST_URI']
    'kafka://host/ok?mode=hot'
    """
    global immutable_config_values
    if immutable_config_values is not None:
        raise Exception('load() function should only be called once.')
    conf = {k.upper(): v for k, v in parse_yaml_file(fname).items()} if fname \
        else {}
    str_conf = {k: v for k, v in conf.items() if type(v) is str}
    str_conf.update(opts)
    opts2env(str_conf, prefix)
    conf.update(load_from_env(prefix))
    immutable_config_values = OrderedDict(conf)
    return conf


def parse_yaml_file(fname: str) -> Dict[str, Any]:
    doc = open(fname, 'r')
    return yaml.full_load(doc)


def get(key=None, default=None) -> Union[ConfigStore, ConfigValue]:
    if immutable_config_values is None:
        raise Exception('Config not yet loaded. Call load() instead.')
    if key:
        val = immutable_config_values.get(key.upper(), default)
        return val.copy() if hasattr(val, 'copy') else val  # type: ignore
    else:
        return immutable_config_values.copy()


def factory(params: Dict) -> Any:
    if params.get('package'):
        params['module'] = '.'.join([params['package'], params['module']])
    mod = importlib.import_module(params['module'])
    cl = params['class']
    del(params['package'])
    del(params['module'])
    del(params['class'])
    return getattr(mod, cl)(**params)


if __name__ == "__main__":
    import doctest
    doctest.testmod()
